% Power and channel resource allocation for F/TDMA relay network with
% optimized channel resource allocation among different control nodes
% Direct transmission (DT) and decode and forward (DF)

clc;
clear;
close all;

% Configuration for system topology (with relay)
path_factor = 4;
node_pos = [ 0 ; 0+sqrt(-1)*5 ; 5 ; 5+sqrt(-1)*5 ; 1.5+sqrt(-1)*1.5 ; 4+sqrt(-1)*4 ];
P_max_dB = [ 25 ; 25 ; 25 ; 25 ; 25 ; 25 ];
M = [ 1 3 ; 2 1 ; 3 2 ; 2 4 ; 4 3 ];
% Enable different links to specify the topology of the network
DT_enable = [ 1 ; 1 ; 1 ; 1 ; 1 ];
RDF_enable = [ 0 0 0 0 1 1 ; 0 0 0 0 1 1 ; 0 0 0 0 1 1 ; 0 0 0 0 1 1 ; 0 0 0 0 1 1 ];
% Control nodes for all the links
C = [ 1 ; 1 ; 2 ; 3 ; 3 ];
% Channel resource of each control node
beta = [ 0.4 ; 0.2 ; 0.4 ];
% beta = [ 1/3 ; 1/3 ; 1/3 ];
% Minimum channel allocation
T_min = 0.01;

% % Configuration for system topology (without relay)
% path_factor = 4;
% node_pos = [ 0 ; 0+sqrt(-1)*5 ; 5 ; 5+sqrt(-1)*5 ];
% P_max_dB = [ 15 ; 15 ; 15 ; 15  ];
% M = [ 1 3 ; 2 1 ; 3 2 ; 2 4 ; 4 3 ];
% % Enable different links to specify the topology of the network
% DT_enable = [ 1 ; 1 ; 1 ; 1 ; 1 ];
% RDF_enable = [ 0 0 0 0 ; 0 0 0 0 ; 0 0 0 0 ; 0 0 0 0 ; 0 0 0 0 ];
% % Control nodes for all the links
% C = [ 1 ; 1 ; 2 ; 3 ; 3 ];
% % Channel resource of each control node
% beta = [ 1/3 ; 1/3 ; 1/3 ];
% % Minimum channel allocation
% T_min = 0.01;

% Configuration for optimization parameters
step_dual_P = 5e-5;  % step size in updating dual variables for power
c = 1e-4;            % coefficients in quadratic terms
Num_JointIter = 5;  % number of iterations in distributed joint-power-channel resource allocation
Num_AuxilIter = 2;
step_beta = 30e-2;    % step size in updating channel resources of control nodes
MAX_ITER = 100;       % maximum number of iterations for updating channel resources of control nodes

% Calculate channel gains
N=length(node_pos);
S = unique(M(:,1));
D = unique(M(:,2));
tmp = 1:N;
tmp([S;D])=0;
R = find(tmp);
P_max = 10.^(P_max_dB./10);
Num_Link = size(M,1);
% g_m^{sd}: power attenuation from source to destination, |M|-by-1
gsd = abs(node_pos(M(:,1))-node_pos(M(:,2))).^(-path_factor);
gsd_tmp = gsd*ones(1,N);
% g_m^{sr}: power attenuation from source to relay, |M|-by-(|N|+|J|)
gsr = abs(repmat(node_pos(M(:,1)),[1 length(node_pos)])-repmat(node_pos.',[Num_Link,1]));
gsr(gsr==0) = 1e6;
gsr = gsr.^(-path_factor);
% gsr(gsr==1e6^(-path_factor)) = 0;
% g_m^{rd}: power attenuation from relay to destination, |M|-by-(|N|+|J|)
grd = abs(repmat(node_pos(M(:,2)),[1 length(node_pos)])-repmat(node_pos.',[Num_Link,1]));
grd(grd==0) = 1e6;
grd = grd.^(-path_factor);
% grd(grd==1e6^(-path_factor)) = 0;

% Initial power allocation
% All the initial values are set feasible to speed up the convergence rate
Pmk = zeros(Num_Link, 1);
Psk = zeros(Num_Link, N);
Prk = zeros(Num_Link, N);
for index_node=1:1:length(S)
    % Calculate the number of possible power allocations of the current node
    tmp_num = sum(DT_enable(M(:,1)==S(index_node))) + sum(sum(RDF_enable(M(:,1)==S(index_node),:),2),1);
    for tmp_index=1:1:Num_Link
        if ((M(tmp_index,1)~=S(index_node))&&(M(tmp_index,2)~=S(index_node)))
            tmp_num = tmp_num + RDF_enable(tmp_index, S(index_node)-(M(tmp_index,1)<S(index_node))-(M(tmp_index,2)<S(index_node)));
        end
    end
    % Power allocation of the source nodes for direct transmission
    Pmk(M(:,1)==S(index_node)) = P_max(S(index_node),1)/tmp_num;
    % Power allocation of the source nodes for relay transmission
    Psk(M(:,1)==S(index_node),:) = P_max(S(index_node),1)/tmp_num;
    % Power allocation of the relay nodes for relay transmission
    Prk(:,S(index_node)) = P_max(S(index_node),1)/tmp_num;
end
Pm = Pmk;
Ps = Psk;
Pr = Prk;

% Initial channel allocation
% Qm: |M|-by-1; Qr: |M|-by-(|N|+|J|)
Qm = zeros(Num_Link,1);
Qr = zeros(Num_Link,N);
for index_control=1:1:length(beta)
    tmp_link = (1:1:Num_Link)'.*(C==index_control);
    tmp_link = tmp_link(tmp_link>0);
    num_channel = sum(C==index_control);
    for index_m=1:1:length(tmp_link)
        tmp_m = tmp_link(index_m);
        num_channel = num_channel + sum(gsr(tmp_m,RDF_enable(tmp_m,:)==1)>=gsd(tmp_m));
    end
    Qm(C==index_control,1) = beta(index_control)/num_channel;
    Qr(C==index_control,:) = beta(index_control)/num_channel;
end
% If gsr<=gsr for a link with certain relay, that relay will not be used
for m = 1:Num_Link
    for j = 1:N
        if ((RDF_enable(m,j)==0)||(gsr(m,j)<=gsd(m)))
            Qr(m,j) = 0;
        end
    end
end

dual_P = 1e-4*ones(N,1);
dual_T = ones(length(beta),1);

prev_Pm = -ones(Num_Link,1);
prev_Ps = -ones(Num_Link,N);
prev_Pr = -ones(Num_Link,N);
prev_Qm = -ones(Num_Link,1);
prev_Qr = -ones(Num_Link,N);
epsilon = 1e-8;
iter = 1;
all_Pm = zeros(Num_Link, MAX_ITER);
all_Ps = zeros(Num_Link*N, MAX_ITER);
all_Pr = zeros(Num_Link*N, MAX_ITER);
all_Qm = zeros(Num_Link, MAX_ITER);
all_Qr = zeros(Num_Link*N, MAX_ITER);
all_beta = zeros(length(beta), MAX_ITER);
%all_Rate = zeros(Num_Link, MAX_ITER*Num_JointIter);
all_Rate = zeros(Num_Link, 0);
%all_PowerFeasible = zeros(N, MAX_ITER*Num_JointIter);
all_PowerFeasible = zeros(N,0);
%iter_num = 1;
while ((norm([Pm Ps Pr Qm Qr]-[prev_Pm prev_Ps prev_Pr prev_Qm prev_Qr])/norm([prev_Pm prev_Ps prev_Pr prev_Qm prev_Qr])>epsilon)&&(iter<=MAX_ITER))
    display(iter)
    
    prev_Pm = Pm;
    prev_Ps = Ps;
    prev_Pr = Pr;
    prev_Qm = Qm;
    prev_Qr = Qr;
    [Rate_Joint, Pm, Ps, Pr, dual_P, Qm, Qr, dual_T Power_Feasible] = power_channel_df_traditional_proximalpoint(M, N, C, RDF_enable, gsd, gsr, grd, beta, T_min, Num_JointIter, step_dual_P, c, P_max, Pmk, Psk, Prk, dual_P, Qm, Qr, dual_T);
    Pmk = Pm;
    Psk = Ps;
    Prk = Pr;
% beta = max(beta + step_beta*dual_T, T_min);
% channel adjustment
% if mod(iter,10)==0
    cvx_begin quiet
    variable x(3,1)
    minimize norm(x-beta-step_beta*dual_T)
    x(1)>=4*T_min;
    x(2)>=2*T_min;
    x(3)>=4*T_min;
    sum(x)<=1;
    cvx_end
% end
    beta = x;
    
    all_Pm(:, iter) = Pm;
    all_Ps(:, iter) = Ps(:);
    all_Pr(:, iter) = Pr(:);
    all_Qm(:, iter) = Qm;
    all_Qr(:, iter) = Qr(:);
    all_beta(:, iter) = beta;
    %all_Rate(:,(iter-1)*Num_JointIter*Num_AuxilIter+1:iter*Num_JointIter*Num_AuxilIter) = Rate_Joint;
    all_Rate = [all_Rate Rate_Joint];
    %all_PowerFeasible(:,(iter-1)*Num_JointIter*Num_AuxilIter+1:iter*Num_JointIter*Num_AuxilIter) = Power_Feasible;
    all_PowerFeasible = [ all_PowerFeasible Power_Feasible];
    iter = iter + 1;
end

% display(Power_Feasible(:,end))

figure;
y = sum(all_Rate,1);
semilogy(abs(y-y(end)))
xlabel('Number of iterations')
ylabel('Gap from optimum')

% figure;
% interval = 10;
% plot([1:interval:iter-1 iter-1], all_beta(1,[1:interval:iter-1 iter-1]),'o-')
% hold on
% plot([1:interval:iter-1 iter-1], all_beta(2,[1:interval:iter-1 iter-1]),'s-')
% plot([1:interval:iter-1 iter-1], all_beta(3,[1:interval:iter-1 iter-1]),'*-')
% legend('User 1', 'Relay 1', 'Relay 2')
% xlabel('Number of channel resources adjustments')
% ylabel('Channel resource fraction')